/**
 * Copyright (C) 2010-2012 Agrosense info@agrosense.eu
 *
 * Licensed under the Eclipse Public License - v 1.0 (the "License"); you may
 * not use this file except in compliance with the License. You may obtain a
 * copy of the License at
 *
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 * License for the specific language governing permissions and limitations under
 * the License.
 */
package nl.cloudfarming.client.isobus.model;

import java.io.*;

/**
 * Class that reads LITTLE endian binary files
 *
 * @author Gerben Feenstra & Merijn Zengers
 */
public class BinaryReader implements Closeable {
    
    private InputStream inputStream = null;

    public BinaryReader(InputStream inputStream) {
        this.inputStream = inputStream;
    }

    /**
     * Method that reads from the binairyFile given to the constructor Reads the
     * number of bytes specified in XmlField.BinaryDataType Will reverse the
     * bytes to get Big endian
     *
     * @param field The field to read
     * @return value that has been red
     */
    public long read(XmlField field) throws IOException {
        assert inputStream != null;
        BinaryDataType datatype = field.getDataType();
        byte[] bytes = new byte[datatype.getSize()];
        inputStream.read(bytes);
        return convert(bytes, datatype);
    }
        
    public static long convert(byte[] bytes, BinaryDataType datatype) throws IOException {
        long value = 0;
        ByteArrayInputStream bais = new ByteArrayInputStream(bytes);
        DataInputStream dataInputStream = new DataInputStream(bais);
        if (dataInputStream.available() == datatype.getSize()) {
            switch (datatype) {
                case BIT8_SIGNED:
                    value = dataInputStream.readByte();
                    break;
                case BIT8_UNSIGNED:
                    value = dataInputStream.readUnsignedByte();
                    break;
                case BIT16_SIGNED:
                    Short signed16temp = dataInputStream.readShort();
                    value = (long) Short.reverseBytes(signed16temp);
                    break;
                case BIT16_UNSIGNED:
                    Short unsigned16temp = dataInputStream.readShort();
                    value = (long) Short.reverseBytes(unsigned16temp);
                    if (value < 0) {
                        value += 65_536; // 2^16
                    }
                    break;
                case BIT32_SIGNED:
                    Integer signed32temp = dataInputStream.readInt();
                    value = (long) Integer.reverseBytes(signed32temp);
                    break;
                case BIT32_UNSIGNED:
                    Integer unsigned32temp = dataInputStream.readInt();
                    value = (long) Integer.reverseBytes(unsigned32temp);
                    if (value < 0) {
                        value += 4_294_967_296L; // 2^32
                    }
                    break;
            }
        }
        return value;
    }

    @Override
    public void close() throws IOException {
        inputStream.close();
        inputStream = null;
    }

    /**
     *
     * @return True if the inputstream != null and has still bits available.
     * @throws IOException
     */
    public boolean available() throws IOException {
        return inputStream != null && inputStream.available() > 0;
    }
}
