/*
 * Copyright (c) 2012 AgroSense.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * Contributors:
 *    AgroSense - initial API and implementation and/or initial documentation
 */
package nl.cloudfarming.client.isobus.model;

import java.util.Arrays;
import nl.cloudfarming.client.util.UOM;

/**
 * Enum which describes all the DDI as formulated in iso 11783-11
 *
 * latest version is available at http://dictionary.isobus.net/isobus/
 *
 * @author johan
 */
public enum DDI {

    /**
     * This DDE is used to specify which version of the Data Dictionary is being
     * used. Range 0 - 2147483647
     */
    DDI_0("This DDE is used to specify which version of the Data Dictionary is being used.", UOM.NOT_SPECIFIED, new byte[]{0, 0}, 1),
    /**
     * Setpoint Application Rate specified as volume per area Range 0 -
     * 2147483647
     */
    DDI_1("Setpoint Volume Per Area Application Rate", UOM.MM3_PER_M2, new byte[]{0, 1}, 0.01),
    /**
     * Actual Application Rate specified as volume per area Range 0 - 2147483647
     */
    DDI_2("Actual Volume Per Area Application Rate", UOM.MM3_PER_M2, new byte[]{0, 2}, 0.01),
    /**
     * Default Application Rate specified as volume per area use when missing
     * Position data or outside any Treatment Zone Range 0 - 2147483647
     */
    DDI_3("Default Volume Per Area Application Rate", UOM.MM3_PER_M2, new byte[]{0, 3}, 0.01),
    /**
     * Minimum Application Rate specified as volume per area supplied by device
     * as physical minimum 0 - 2147483647
     */
    DDI_4("Minimum Volume Per Area Application Rate", UOM.MM3_PER_M2, new byte[]{0, 4}, 0.01),
    /**
     * Maximum Application Rate specified as volume per area supplied by device
     * as physical maximum Range 0 - 2147483647
     */
    DDI_5("Maximum Volume Per Area Application Rate", UOM.MM3_PER_M2, new byte[]{0, 5}, 0.01),
    /**
     * Setpoint Application Rate specified as mass per area Range 0 - 2147483647
     */
    DDI_6("Setpoint Mass Per Area Application Rate", UOM.MG_PER_M2, new byte[]{0, 6}, 1),
    /**
     * Actual Application Rate specified as mass per area Range 0 - 2147483647
     */
    DDI_7("Actual Mass Per Area Application Rate", UOM.MG_PER_M2, new byte[]{0, 7}, 1),
    /**
     * Default Application Rate specified as mass per area. Use when missing
     * Position data or outside any Treatment Zone. Range 0 - 2147483647
     */
    DDI_8("Default Mass Per Area Application Rate", UOM.MG_PER_M2, new byte[]{0, 8}, 1),
    /**
     * Minimum Application Rate specified as mass per area supplied by device as
     * physical minimum. Range 0 - 2147483647
     */
    DDI_9("Minimum Mass Per Area Application Rate", UOM.MG_PER_M2, new byte[]{0, 9}, 1),
    /**
     * Maximum Application Rate specified as mass per area supplied by device as
     * physical maximum. Range 0 - 2147483647
     */
    DDI_10("Maximum Mass Per Area Application Rate", UOM.MG_PER_M2, new byte[]{0, 10}, 1),
    /**
     * Setpoint Application Rate specified as count per area. Range 0 -
     * 2147483647
     */
    DDI_11("Setpoint Count Per Area Application Rate", UOM.AMOUNT_PER_M2, new byte[]{0, 11}, 0.001),
    /**
     * Actual Application Rate specified as count per area. Range 0 - 2147483647
     */
    DDI_12("Actual Count Per Area Application Rate", UOM.AMOUNT_PER_M2, new byte[]{0, 12}, 0.001),
    /**
     * Default Count Per Area Application Rate Default Application Rate
     * specified as count per area. Comment use when missing Position data or
     * outside any Treatment Zone Range 0 - 2147483647
     */
    DDI_13("Default Count Per Area Application Rate", UOM.AMOUNT_PER_M2, new byte[]{0, 13}, 0.001),
    /**
     * Minimum Application Rate specified as count per area. Comment: supplied by
     * device as physical minimum.
     * Range 0 - 2147483647
     */
    DDI_14("Minimum CountPer Area Application Rate", UOM.AMOUNT_PER_M2, new byte[]{0, 14}, 0.001),
    /**
     * Maximum Application Rate specified as count per area. 
     * Comment: supplied by device as physical maximum.
     * Range 0 - 2147483647
     */
    DDI_15("Maximum Count Per Area Application Rate", UOM.AMOUNT_PER_M2, new byte[]{0, 15}, 0.001),
    /**
     * Setpoint Application Rate specified as distance: e.g. seed spacing of a precision seeder.
     * Range 0 - 2147483647
     */
    DDI_16("Setpoint Spacing Application Rate", UOM.MM, new byte[]{0, 16}, 1),
    /**
     * Actual Application Rate specified as distance: e.g. seed spacing of a precision seeder.
     * Range 0 - 2147483647
     */
    DDI_17("Actual Spacing Application Rate", UOM.MM, new byte[]{0, 17}, 1),
    /**
     * Default Application Rate specified as distance: e.g. seed spacing of a precision seeder.
     * Comment: use when missing Position data or outside any Treatment Zone.
     * Range 0 - 2147483647
     */
    DDI_18("Default Spacing Application Rate", UOM.MM, new byte[]{0, 18}, 1),
    /**
     * Minimum Application Rate specified as distance: e.g. seed spacing of a precision seeder.
     * Comment: supplied by device as physical minimum.
     * Range 0 - 2147483647
     */
    DDI_19("Minimum Spacing Application Rate", UOM.MM, new byte[]{0, 19}, 1),
    /**
     * Maximum Application Rate specified as distance: e.g. seed spacing of a precision seeder.
     * Comment: supplied by device as physical maximum.
     * Range 0 - 2147483647
     */
    DDI_20("Maximum Spacing Application Rate", UOM.MM, new byte[]{0, 20}, 1),
    /**
     * Setpoint Application Rate specified as volume per volume.
     * Range 0 - 2147483647
     */
    DDI_21("Setpoint Volume Per Volume Application Rate", UOM.MM3_PER_M3, new byte[]{0, 21}, 1),
    /**
     * Actual Application Rate specified as volume per volume.
     * Range 0 - 2147483647
     */
    DDI_22("Actual Volume Per Volume Application Rate", UOM.MM3_PER_M3, new byte[]{0, 22}, 1),
    /**
     * Default Application Rate specified as volume per volume.
     * Comment: use when missing Position data or outside any Treatment Zone.
     * Range 0 - 2147483647
     */
    DDI_23("Default Volume Per Volume Application Rate", UOM.MM3_PER_M3, new byte[]{0, 23}, 1),
    /**
     * Minimum Application Rate specified as volume per volume.
     * Comment: supplied by device as physical minimum.
     * Range 0 - 2147483647
     */
    DDI_24("Minimum Volume Per Volume Application Rate", UOM.MM3_PER_M3, new byte[]{0, 24}, 1),
    /**
     * Maximum Application Rate specified as volume per volume.
     * Comment: supplied by device as physical maximum.
     * Range 0 - 2147483647
     */
    DDI_25("Maximum Volume Per Volume Application Rate", UOM.MM3_PER_M3, new byte[]{0, 25}, 1),
    /**
     * Setpoint Application Rate specified as mass per mass.
     * Range 0 - 2147483647
     */
    DDI_26("Setpoint Mass Per Mass Application Rate", UOM.MG_PER_KG, new byte[]{0, 26}, 1),
    /**
     * Actual Application Rate specified as mass per mass.
     * Range 0 - 2147483647
     */
    DDI_27("Actual Mass Per Mass Application Rate", UOM.MG_PER_KG, new byte[]{0, 27}, 1),
    /**
     * Default Application Rate specified as mass per mass.
     * Comment: use when missing Position data or outside any Treatment Zone.
     * Range 0 - 2147483647
     */
    DDI_28("Default Mass Per MassA pplication Rate", UOM.MG_PER_KG, new byte[]{0, 28}, 1),
    /**
     * Minimum Application Rate specified as mass per mass.
     * Comment: supplied by device as physical minimum.
     * Range 0 - 2147483647
     */
    DDI_29("Minimum Mass Per Mass Application Rate", UOM.MG_PER_KG, new byte[]{0, 29}, 1),
    /**
     * Maximum Application Rate specified as mass per mass.
     * Comment: supplied by device as physical maximum.
     * Range 0 - 2147483647
     */
    DDI_30("MaximumMass Per Mass Application Rate", UOM.MG_PER_KG, new byte[]{0, 30}, 1),
    /**
     * Setpoint Application Rate specified as volume per mass.
     * Range 0 - 2147483647
     */
    DDI_31("Setpoint Volume Per Mass Application Rate", UOM.MM3_PER_KG, new byte[]{0, 31}, 1),
    /**
     * Actual Application Rate specified as volume per mass.
     * Range 0 - 2147483647
     */
    DDI_32("Actual Volume Per Mass Application Rate", UOM.MM3_PER_KG, new byte[]{0, 32}, 1),
    /**
     * Default Application Rate specified as volume per mass.
     * Comment: use when missing Position data or outside any Treatment Zone.
     * Range 0 - 2147483647
     */
    DDI_33("Default Volume Per Mass Application Rate", UOM.MM3_PER_KG, new byte[]{0,33}, 1),
    /**
     * Minimum Application Rate specified as volume per mass.
     * Comment: supplied by device as physical minimum.
     * Range 0 - 2147483647
     */
    DDI_34("Minimum Volume Per Mass Application Rate", UOM.MM3_PER_KG, new byte[]{0,34}, 1),
    /**
     * Maximum Application Rate specified as volume per mass.
     * Comment: supplied by device as physical maximum.
     * Range 0 - 2147483647
     */
    DDI_35("Maximum Volume Per Mass Application Rate", UOM.MM3_PER_KG, new byte[]{0,35}, 1),
    /**
     * Setpoint Application Rate specified as volume per time.
     * Range 0 - 2147483647
     */
    DDI_36("Setpoint Application Rate specified as volume per time", UOM.MM3_PER_SECOND, new byte[]{0,36}, 1),
    /**
     * Actual Application Rate specified as volume per time.
     * Range 0 - 2147483647
     */
    DDI_37("Actual Volume Per Time Application Rate", UOM.MM3_PER_SECOND, new byte[]{0,37}, 1),
    /**
     * Default Application Rate specified as volume per time.
     * Comment: use when missing Position data or outside any Treatment Zone.
     * Range 0 - 2147483647
     */
    DDI_38("Default Volume Per Time Application Rate", UOM.MM3_PER_SECOND, new byte[]{0,38}, 1),
    /**
     * Minimum Application Rate specified as volume per time.
     * Comment: supplied by device as physical minimum.
     * Range 0 - 2147483647
     */
    DDI_39("Minimum Volume Per Time Application Rate", UOM.MM3_PER_SECOND, new byte[]{0,39}, 1),
    /**
     * Maximum Application Rate specified as volume per time.
     * Comment: supplied by device as physical maximum.
     * Range 0 - 2147483647
     */
    DDI_40("Maximum Volume Per Time Application Rate", UOM.MM3_PER_SECOND, new byte[]{0,40}, 1),
    /**
     * Setpoint Application Rate specified as mass per time.
     * Range 0 - 2147483647
     */
    DDI_41("Setpoint Mass Per Time Application Rate", UOM.MG_PER_SECOND, new byte[]{0,41}, 1),
    /**
     * Actual Application Rate specified as mass per time.
     * Range 0 - 2147483647
     */
    DDI_42("Actual Mass Per Time Application Rate", UOM.MG_PER_SECOND, new byte[]{0,42}, 1),
    /**
     * Default Application Rate specified as mass per time.
     * Comment: use when missing Position data or outside any Treatment Zone.
     * Range 0 - 2147483647
     */
    DDI_43("Default Mass Per Time Application Rate", UOM.MG_PER_SECOND, new byte[]{0,43}, 1),
    /**
     * Minimum Application Rate specified as mass per time.
     * Comment: supplied by device as physical minimum.
     * Range 0 - 2147483647
     */
    DDI_44("Minimum Mass Per Time Application Rate", UOM.MG_PER_SECOND, new byte[]{0,44}, 1),
    /**
     * Maximum Application Rate specified as mass per time.
     * Comment: supplied by device as physical maximum.
     * Range 0 - 2147483647
     */
    DDI_45("Maximum Mass Per Time Application Rate", UOM.MG_PER_SECOND, new byte[]{0,45}, 1),
    /**
     * Setpoint Application Rate specified as count per time.
     * Range 0 - 2147483647
     */
    DDI_46("Setpoint Count Per Time Application Rate", UOM.AMOUNT_PER_SECOND, new byte[]{0,46}, 0.001),
    /**
     * Actual Application Rate specified as count per time.
     * Range 0 - 2147483647
     */
    DDI_47("Actual Count Per Time Application Rate", UOM.AMOUNT_PER_SECOND, new byte[]{0,47}, 0.001),
    /**
     * Default Application Rate specified as count per time.
     * Comment: use when missing Position data or outside any Treatment Zone.
     * Range 0 - 2147483647
     */
    DDI_48("Default Count Per Time Application Rate", UOM.AMOUNT_PER_SECOND, new byte[]{0,48}, 0.001),
    /**
     * Minimum Application Rate specified as count per time.
     * Comment: supplied by device as physical minimum.
     * Range 0 - 2147483647
     */
    DDI_49("Minimum Count Per Time Application Rate", UOM.AMOUNT_PER_SECOND, new byte[]{0,49}, 0.001),
    /**
     * Maximum Application Rate specified as count per time.
     * Comment: supplied by device as physical maximum.
     * Range 0 - 2147483647
     */
    DDI_50("Maximum Count Per Time Application Rate", UOM.AMOUNT_PER_SECOND, new byte[]{0,50}, 0.001),
    /**
     * Setpoint Tillage Depth of Device Element below soil surface, value
     * increases with depth.
     * Range 0 - 2147483647
     */
    DDI_51("Setpoint Tillage Depth", UOM.MM, new byte[]{0,51}, 1),
    /**
     * Actual Tillage Depth of Device Element below soil surface, value
     * increases with depth.</br>
     * Range 0 - 2147483647
     */
    DDI_52("Actual Tillage Depth", UOM.MM, new byte[]{0,52}, 1),
    /**
     * Default Tillage Depth of Device Element below soil surface, value
     * increases with depth.</br>
     * Comment: use when missing Position data or outside any Treatment Zone.
     * Range 0 - 2147483647
     */
    DDI_53("Default Tillage Depth", UOM.MM, new byte[]{0,53}, 1),
    
    /**
     * Actual Working Width of Device Element.
     * Comment: This is the effective / active working width during operation.
     * Range 0 - 2147483647
     */
    DDI_67("Actual Working Width", UOM.MM, new byte[]{0, 67}, 1),
    /**
     * Accumulated Application specified as mass.
     * Comment: is a counter of a device element.
     * Range 0 - 2147483647
     */
    DDI_81("Application Total Mass", UOM.KG, new byte[]{0, 81}, 1),
    /**
     * Accumulated Area.
     * Comment: is a counter of a machine element.
     * Range 0 - 2147483647
     */
    DDI_116("Total Area", UOM.M2, new byte[]{0, 116}),
    /**
     * Accumulated Distance in working position.
     * Comment: is a counter of a machine element.
     * Range 0 - 2147483647
     */
    DDI_117("Effective Total Distance", UOM.MM, new byte[]{0, 117}, 1),
    /**
     * Accumulated Distance out of working position.
     * Comment: is a counter of a machine element.
     * Range 0 - 2147483647
     */
    DDI_118("Ineffective Total Distance", UOM.MM, new byte[]{0, 118}, 1),
    /**
     * Definition Accumulated Time in working position.
     * Comment: is a counter of a machine element.
     * Range 0 - 2147483647
     */
    DDI_119("Effective Total Time", UOM.SECOND, new byte[]{0, 119}, 1),
    /**
     * Accumulated Time out of working position.
     * Comment: is a counter of a machine element.
     * Range 0 - 2147483647
     */
    DDI_120("Ineffective Total Time", UOM.SECOND, new byte[]{0, 120}, 1),
    /**
     * X direction offset of a DeviceElement relative to a Device. Range
     * -2147483648 - 2147483647
     */
    DDI_134("Device Element Offset X", UOM.MM, new byte[]{0, Byte.parseByte("-122")}, 1),
    /**
     * Y direction offset of a DeviceElement relative to a Device. Range
     * -2147483648 - 2147483647
     */
    DDI_135("Device Element Offset Y", UOM.MM, new byte[]{0, Byte.parseByte("-121")}, 1),
    /**
     * Actual Work State, 2 bits defined as 00=disabled/off, 01=enabled/on,
     * 10=error, 11=undefined/not installed. See the DDI 289 attachment
     * "ISO11783-11-DDI-289-SetpointWorkState" for implementation guidelines.
     * This DDE has been revised in 2012 to be used as the Actual Work State. A
     * separate Setpoint Work State was added to the data dictionary at that
     * time. Original comment: this DDE was a result of March 2005 TF10 meeting.
     * Range 0-3
     */
    DDI_141("Actual Work State", UOM.NOT_SPECIFIED, new byte[]{0, Byte.parseByte("-115")}, 1), // AKA on/off
    /**
     * The Setpoint Value Latency Time is the time lapse between the moment of
     * receival of a setpoint value command by the working set and the moment
     * this setpoint value is physically applied on the device. That means if
     * the setpoint value is communicated on the network (CAN bus) but the
     * system needs 2 seconds to adjust the value physically on the desired unit
     * (device element) then the Setpoint Latency Time is 2 seconds. The
     * setpoint time latency value can only be positive.</br> The use of this
     * DDE is to inform the overall system (e.g. Dektop Software, Task
     * Controller) how the system works. The TC shall not shift this information
     * into log files nor shall the device do that when sending actual values.
     *
     */
    DDI_142("Physical Setpoint Time Latency", UOM.MS, new byte[]{0, Byte.parseByte("-114")}, 1),
    /**
     * Specifies the actual state of section control. The value definitions are:
     * Byte 1 (bitfield) Bit 0-1: 00 = manual/off, 01 = auto/on, 10 = error
     * indicator, 11 = undefined/not installed. Bits 2-7: reserved, set to 0.
     * Bytes 2-4: reserved, set to 0. See attachment for more information</br>
     * In section control systems, the section control master and its clients
     * need to be synchronized in terms of their general state or activation by
     * the user (System activated/deactivated in individual setups). This DDE
     * serves 2 purposes, one is to synchronize the section control state and
     * the other is to enable a TC client to announce the support and initial
     * state of its section control capabilities. Synchronisation by the section
     * control master (TC) is done by setting the section control state of
     * connected TC clients with a process data set value message with this DDE.
     * It is recommended that TC clients reply their state immediately (within
     * 250 ms) when such a message is received. The property flag "setable" and
     * the trigger method "on change" should be used with this DDE. The state
     * "manual/off" indicates that the device is in manual state and will ignore
     * all control commands for section control. The "auto/on" state indicates
     * that the client accepts the section control commands as far as its
     * overall process state allows. Listed below are 4 example Use Cases for
     * this DDE: Use case "Start up operation": 1. During a start up the
     * implement shall set the SCS to ‘manual mode’. Use case "Auto request from
     * TC": 1. The implement receives an ‘auto’ request from TC. 2. The
     * implement shall check whether all setup conditions are fulfilled to allow
     * section control. 3. If this check is ok: The implement may respond with
     * ‘auto mode’ and set its internal SCS client to ‘auto mode’. 4. If this
     * check not ok: The implement shall respond with ‘manual mode’. The
     * internal state is still in manual mode. The TC may inform the operator
     * accordingly. Use case "Manual request from TC": 1. The implement receives
     * a manual request from TC. 2. If the implement is still in ‘auto mode’ the
     * implement shall set its internal SCS client to ‘manual mode’. Use case
     * "Loss of requirements for auto mode": 1. The implement internal setup
     * conditions don’t allow for automatic section control anymore. 2. The
     * implement shall set the internal SCS client to ‘manual mode’. 3. The
     * implement shall send the SCS to inform the TC accordingly. On reception
     * of this ‘manual mode’ the TC/Section Control Master may inform the
     * operator accordingly.
     *
     */
    DDI_160("Section Control State", UOM.NOT_SPECIFIED, new byte[]{0, Byte.parseByte("-96")}, 1),
    /**
     * Combination of the actual work states of individual sections or units
     * (e.g. nozzles) number 1 to 16 into a single actual work state of their
     * parent device element. The actual condensed work state contains the
     * child element actual work states, in the driving direction from left to
     * right, where the leftmost child element actual work state are the 2
     * lowest significant bits of the Process Data Value. Each child device
     * elements actual work state is represented by 2 bits and defined as:
     * 00 = disabled/off, 01 = enabled/on, 10 = error indicator, 11 =
     * undefined/not installed. In total 16 child device element actual work
     * states can be contained in one actual condensed work state of their
     * parent device element. If less than 16 child device element actual
     * work states are available, then the unused bits shall be set to value
     * 11 (not installed). When the parent device element contains the
     * Actual Condensed Work State DDE, then the device descriptor shall
     * not contain the individual actual work state DDEs of the child device
     * elements.
     * 
     */
    DDI_161("Actual Condensed Work State", UOM.NOT_SPECIFIED, new byte[]{0, Byte.parseByte("-95")}, 1),
    /**
     * Request Default Process Data. This DDE is the highest ISO assigned
     * entity. The range above this number is reserved for manufacture specific
     * DDE's. Range 0-0
     */
    DDI_57343("Request Default Process Data", UOM.NOT_SPECIFIED, new byte[]{Byte.parseByte("-33"), Byte.parseByte("-128")}, 1);
    // TODO: add more DDI values...
    public static final double DEFAULT_RESOLUTION = 1.0;
    private final String description;
    private final byte[] DDEntityId; // length 2
    private final UOM uom; //todo: to enum; maybe later: generic data conversion based on input value and from/to UOMs.
    private final double resolution;

    private DDI(String description, UOM uom, byte[] DDEntityId, double resolution) {
        this.DDEntityId = DDEntityId;
        this.description = description;
        this.uom = uom;
        this.resolution = resolution;
    }

    private DDI(String description, UOM uom, byte[] DDEntityId) {
        this(description, uom, DDEntityId, DEFAULT_RESOLUTION);
    }

    public String getDescription() {
        return description;
    }

    public byte[] getDDEntityId() {
        return DDEntityId;
    }

    public UOM getUom() {
        return uom;
    }

    public double getResolution() {
        return resolution;
    }

    public int toExportValue(double value) {
        return (int) (value / resolution);
    }

    public double toImportValue(int value) {
        return (((double) value) * resolution);
    }

    @Override
    public String toString() {
        String result = description;
        if (uom != null) {
            result += ", " + uom.getLabel();
        }
        return result;
    }

    public static DDI findById(byte[] bytes) {
        for (DDI ddi : DDI.values()) {
            if(Arrays.equals(bytes, ddi.getDDEntityId())) {
                return ddi;
            }
        }
        return null;
    }

}
