/**
 * Copyright (C) 2010-2012 Agrosense info@agrosense.eu
 *
 * Licensed under the Eclipse Public License - v 1.0 (the "License"); you may
 * not use this file except in compliance with the License. You may obtain a
 * copy of the License at
 *
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 * License for the specific language governing permissions and limitations under
 * the License.
 */
package nl.cloudfarming.client.isobus.model.util;

import java.util.logging.Level;
import java.util.logging.Logger;
import nl.cloudfarming.client.field.model.TreatmentZone;
import nl.cloudfarming.client.isobus.model.*;

/**
 * Can create a grid
 *
 * @author Merijn Zengers
 */
public class IsobusObjectFactoryUtil {

    private static final Logger LOGGER = Logger.getLogger("nl.cloudfarming.client.isobus.model.util");

    /**
     * Create an IsobusTreatmentZone using an Agrosense TreatmentZone and a
     * TreatmentzoneCode for the task in the context of this treatmentzone. Will
     * add a ProcessDataVariable to the created IsobusTreatmentZone Will set the
     * processDataDDI to hex 8D as provided in ISO_11783-11
     *
     * @param presentation may be null
     * @param treatmentZone Agrosense Treatmentzone
     * @param treatmentZoneCode Code
     * @return Isobus Treatmentzone
     */
    public static IsobusTreatmentZone createIsobusTreatmentZone(ValuePresentation presentation, TreatmentZone treatmentZone, short treatmentZoneCode, DDI ddi, double multiplier) {
        if (treatmentZone.getValue() != null) {
            IsobusTreatmentZone isobusTreatmentZone = new IsobusTreatmentZone();
            isobusTreatmentZone.setTreatmentZoneDesignator(treatmentZone.getName());
            isobusTreatmentZone.setTreatmentZoneCode(treatmentZoneCode);
            isobusTreatmentZone.setTreatmentZoneColor(new Double(treatmentZone.getValue()).shortValue());

            // ProcessDataVariable for this treatmentzone
            ProcessDataVariable processDataVariable = createProcessDataVariable(treatmentZone.getValue(), presentation, ddi, multiplier);
            isobusTreatmentZone.getPDV().add(processDataVariable);
            return isobusTreatmentZone;
        } else {
            LOGGER.log(Level.SEVERE, "Treatment zone has no value set cannot create isobus treatment zone from it {0}", treatmentZone.getName());
        }
        return null;
    }

    /**
     * Create a {@link ProcessDataVariable} with the value and {@link ValuePresentation}
     * passed
     *
     * @param value the value to set
     * @param presentation the valuePresentation to set
     * @param ddi isobus data dictionary id
     * @return the created ProcesDataVariable
     */
    public static ProcessDataVariable createProcessDataVariable(double value, ValuePresentation presentation, DDI ddi, double multiplier) {
        ProcessDataVariable processDataVariable = new ProcessDataVariable();

        processDataVariable.setProcessDataDDI(ddi.getDDEntityId()); // A
        processDataVariable.setProcessDataValue(ddi.toExportValue(value * multiplier)); // B

        if (presentation != null) {
            processDataVariable.setValuePresentationIdRef(presentation);
        }
        return processDataVariable;
    }
    
    
    
}
