/**
 * Copyright (C) 2010-2012 Agrosense info@agrosense.eu
 *
 * Licensed under the Eclipse Public License - v 1.0 (the "License"); you may
 * not use this file except in compliance with the License. You may obtain a
 * copy of the License at
 *
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 * License for the specific language governing permissions and limitations under
 * the License.
 */
package nl.cloudfarming.client.isobus.model.util;

import com.vividsolutions.jts.geom.Coordinate;
import java.awt.geom.Point2D;
import java.util.Arrays;
import nl.cloudfarming.client.isobus.model.DDI;
import nl.cloudfarming.client.isobus.model.DataLogValue;
import nl.cloudfarming.client.isobus.model.DeviceElement;
import nl.cloudfarming.client.isobus.model.IsobusTime;
import nl.cloudfarming.client.lib.geotools.GeometryTools;

/**
 * Util that provides functionality to interpret Timelog data
 *
 * @author Merijn Zengers <Merijn.Zengers@gmail.nl>
 */
public class TimeLogUtil {

    /**
     * Calculates a coordinate from a origin with a x and y offset in
     * millimeters and a bearing in decimal degrees from due north ranging
     * from -180 to 180 where 0 is due north.
     * The X-axis is specified as positive in the normal driving direction.
     * The Y-axis is specified as positive to the right side of the device relative to the normal driving direction.
     *
     * @param origin the origin coordinate in WGS84
     * @param xOffset the x offset in millimeters
     * @param yOffset the y offset in millimeters
     * @param bearing bearing in decimal degrees from due north ranging
     * from -180 to 180 where 0 is due north
     * @return
     */
    public static Coordinate getCoordinate(Coordinate origin, double xOffset, double yOffset, double bearing) {
        double xMetres = xOffset / 1000;
        double yMetres = yOffset / 1000;
        
        //pythagoras a2 + b2 = c2
        double distance = Math.hypot(xMetres, yMetres);
        
        // ~angle of the offset 
        double angle = Math.toDegrees(Math.atan2(yOffset, xOffset)); 
        
        //correct bearing
        //If y < 0 offset is to the left of the device the negative degrees
        //Else offset is to the right of the vehicle
        //Also correct for the -180 to 180 range
        //if initial bearing is -95 and y offset is negative the calculation returns -185 this is outside the range.
        //The calculated value should be +175
        double correction = bearing + angle;
        double correctedBearing = correction  < -180 ? 360 + correction : correction > 180 ? -360 + correction : correction;

        Point2D calculated = GeometryTools.getDestinationPoint(origin.y, origin.x, correctedBearing, distance);
        return new Coordinate(calculated.getX(), calculated.getY());
    }

    /**
     * Get the x offset for a specific deviceElement Searches in the provided {@link IsobusTime}
     * for a {@link DataLogValue} which has a reference to the provided
     * DeviceElement and {@link DataLogValue#processDataDDI} that matches {@link DDI#DDI_134}
     *
     * @param isobusTime
     * @param deviceElement
     * @return the found {@link DataLogValue#processDataValue} times the {@link DDI#resolution}
     * or null if not found
     */
    public static Double getXOffsetForDeviceElement(IsobusTime isobusTime, DeviceElement deviceElement) {
        return getProcessDataValueForDeviceElement(isobusTime, deviceElement, DDI.DDI_134);
    }

    /**
     * Get the y offset for a specific deviceElement Searches in the provided {@link IsobusTime}
     * for a {@link DataLogValue} which has a reference to the provided
     * DeviceElement and {@link DataLogValue#processDataDDI} that matches {@link DDI#DDI_134}
     *
     * @param isobusTime
     * @param deviceElement
     * @return the found {@link DataLogValue#processDataValue} times the {@link DDI#resolution}
     * or null if not found
     */
    public static double getYOffsetForDeviceElement(IsobusTime isobusTime, DeviceElement deviceElement) {
        return getProcessDataValueForDeviceElement(isobusTime, deviceElement, DDI.DDI_135);
    }

    /**
     * Get a {@link DataLogValue#processDataValue} times the {@link DDI#resolution}
     * for a {@link DeviceElement} included in a {@link IsobusTime}
     *
     * @param isobusTime
     * @param deviceElement
     * @param ddi
     * @return the found {@link DataLogValue#processDataValue} times the {@link DDI#resolution}
     * or null if not found
     */
    public static Double getProcessDataValueForDeviceElement(IsobusTime isobusTime, DeviceElement deviceElement, DDI ddi) {
        for (DataLogValue dataLogValue : isobusTime.getDLV()) {
            if (Arrays.equals(ddi.getDDEntityId(), dataLogValue.getProcessDataDDI())) {
                if (deviceElement.equals(dataLogValue.getDeviceElementIdRef())) {
                    return ddi.toImportValue(dataLogValue.getProcessDataValue());
                }
            }
        }
        return null;
    }
}
